@@ -29,6 +29,7 @@ use std::sync::Mutex;
 use std::sync::RwLock;
 use std::sync::Weak;
 use std::sync::atomic::{AtomicI64, AtomicU64, Ordering};
+use std::time;
 
 use tokio::select;
 use tokio::sync::Notify;
@@ -69,8 +70,12 @@ pub type ActorResult = Result<(), Error>;
 /// the actor's mainloop task.
 pub type ActorHandle = tokio::task::JoinHandle<ActorResult>;
 
+/// The type of the "disarm" function returned from [`Activation::prevent_inert_check`].
+pub type DisarmFn = Box<dyn Send + FnOnce()>;
+
 /// A small protocol for indicating successful synchronisation with
 /// some peer; see [Entity::sync].
+#[derive(Debug)]
 pub struct Synced;
 
 /// The core metaprotocol implemented by every object.
@@ -165,6 +170,16 @@ pub trait Entity<M>: Send {
         Ok(())
     }
 
+    /// Optional callback for running actions when the entity's owning [Facet] terminates
+    /// cleanly. Will not be called in case of abnormal shutdown (crash) of an actor.
+    ///
+    /// Programs register an entity's stop hook with [Activation::on_stop].
+    ///
+    /// The default implementation does nothing.
+    fn stop(&mut self, turn: &mut Activation) -> ActorResult {
+        Ok(())
+    }
+
     /// Optional callback for running cleanup actions when the
     /// entity's animating [Actor] terminates.
     ///
@@ -309,6 +324,7 @@ pub struct RunningActor {
     tx: UnboundedSender<SystemMessage>,
     mailbox: Weak<Mailbox>,
     exit_hooks: Vec<Box<dyn Send + FnOnce(&mut Activation, &Arc<ActorResult>) -> ActorResult>>,
+    cleanup_actions: CleanupActions,
     facet_nodes: Map<FacetId, Facet>,
     facet_children: Map<FacetId, Set<FacetId>>,
     root: FacetId,
@@ -338,7 +354,7 @@ pub struct Facet {
     pub facet_id: FacetId,
     /// The ID of the facet's parent facet, if any; if None, this facet is the `Actor`'s root facet.
     pub parent_facet_id: Option<FacetId>,
-    cleanup_actions: CleanupActions,
+    outbound_handles: Set<Handle>,
     stop_actions: Vec<Action>,
     linked_tasks: Map<u64, CancellationToken>,
     inert_check_preventers: Arc<AtomicU64>,
@@ -510,16 +526,19 @@ impl FacetRef {
                             let mut t = Activation::make(&self.actor.facet_ref(state.root),
                                                          Account::new(crate::name!("shutdown")),
                                                          state);
+                            if let Err(err) = t._terminate_facet(t.state.root, exit_status.is_ok()) {
+                                // This can only occur as the result of an internal error in this file's code.
+                                tracing::error!(?err, "unexpected error from terminate_facet");
+                                panic!("Unexpected error result from terminate_facet");
+                            }
+                            // TODO: The linked_tasks are being cancelled above ^ when their Facets drop.
+                            // TODO: We don't want that: we want (? do we?) exit hooks to run before linked_tasks are cancelled.
+                            // TODO: Example: send an error message in an exit_hook that is processed and delivered by a linked_task.
                             for action in std::mem::take(&mut t.state.exit_hooks) {
                                 if let Err(err) = action(&mut t, &exit_status) {
                                     tracing::error!(?err, "error in exit hook");
                                 }
                             }
-                            if let Err(err) = t._terminate_facet(t.state.root, false) {
-                                // This can only occur as the result of an internal error in this file's code.
-                                tracing::error!(?err, "unexpected error from disorderly terminate_facet");
-                                panic!("Unexpected error result from disorderly terminate_facet");
-                            }
                             *g = ActorState::Terminated {
                                 exit_status: Arc::clone(&exit_status),
                             };
@@ -584,15 +603,20 @@ impl<'activation> Activation<'activation> {
     /// Core API: assert `a` at recipient `r`.
     ///
     /// Returns the [`Handle`] for the new assertion.
-    pub fn assert<M: 'static + Send>(&mut self, r: &Arc<Ref<M>>, a: M) -> Handle {
+    pub fn assert<M: 'static + Send + std::fmt::Debug>(&mut self, r: &Arc<Ref<M>>, a: M) -> Handle {
         let handle = next_handle();
-        if let Some(f) = self.active_facet() {
-            f.insert_retract_cleanup_action(&r, handle);
+        if let Some(f) = self.state.get_facet(self.facet.facet_id) {
+            tracing::trace!(?r, ?handle, ?a, "assert");
+            f.outbound_handles.insert(handle);
             drop(f);
+            self.state.insert_retract_cleanup_action(&r, handle);
             {
                 let r = Arc::clone(r);
                 self.pending.queue_for(&r).push(Box::new(
-                    move |t| t.with_entity(&r, |t, e| e.assert(t, a, handle))));
+                    move |t| t.with_entity(&r, |t, e| {
+                        tracing::trace!(?handle, ?a, "asserted");
+                        e.assert(t, a, handle)
+                    })));
             }
         }
         handle
@@ -612,22 +636,33 @@ impl<'activation> Activation<'activation> {
     /// # Panics
     ///
     /// Panics if `r` is not part of the active actor.
-    pub fn assert_for_myself<M: 'static + Send>(&mut self, r: &Arc<Ref<M>>, a: M) -> Handle {
+    pub fn assert_for_myself<M: 'static + Send + std::fmt::Debug>(&mut self, r: &Arc<Ref<M>>, a: M) -> Handle {
         self.immediate_oid(r);
         let handle = next_handle();
         if let Some(f) = self.active_facet() {
-            {
-                let r = Arc::clone(r);
-                f.cleanup_actions.insert(
-                    handle,
-                    CleanupAction::ForMyself(Box::new(
-                        move |t| t.with_entity(&r, |t, e| e.retract(t, handle)))));
-            }
+            tracing::trace!(?r, ?handle, ?a, "assert_for_myself");
+            f.outbound_handles.insert(handle);
             drop(f);
+            {
+                let r = Arc::clone(r);
+                self.state.cleanup_actions.insert(
+                    handle,
+                    CleanupAction::ForMyself(Box::new(
+                        move |t| t.with_entity(&r, |t, e| {
+                            tracing::trace!(?handle, "retracted");
+                            if let Some(f) = t.active_facet() {
+                                f.outbound_handles.remove(&handle);
+                            }
+                            e.retract(t, handle)
+                        }))));
+            }
             {
                 let r = Arc::clone(r);
                 self.pending.for_myself.push(Box::new(
-                    move |t| t.with_entity(&r, |t, e| e.assert(t, a, handle))));
+                    move |t| t.with_entity(&r, |t, e| {
+                        tracing::trace!(?handle, ?a, "asserted");
+                        e.assert(t, a, handle)
+                    })));
             }
         }
         handle
@@ -636,24 +671,28 @@ impl<'activation> Activation<'activation> {
     fn half_link(&mut self, t_other: &mut Activation) {
         let entity_ref = t_other.create::<AnyValue, _>(StopOnRetract);
         let handle = next_handle();
-        self.active_facet().unwrap().insert_retract_cleanup_action(&entity_ref, handle);
+        tracing::trace!(?handle, ?entity_ref, "half_link");
+        self.state.insert_retract_cleanup_action(&entity_ref, handle);
+        self.active_facet().unwrap().outbound_handles.insert(handle);
         t_other.with_entity(&entity_ref, |t, e| e.assert(t, AnyValue::new(true), handle)).unwrap();
     }
 
     /// Core API: retract a previously-established assertion.
     pub fn retract(&mut self, handle: Handle) {
-        if let Some(f) = self.active_facet() {
-            if let Some(d) = f.cleanup_actions.remove(&handle) {
-                self.pending.execute_cleanup_action(d)
-            }
+        if let Some(d) = self.state.cleanup_actions.remove(&handle) {
+            self.pending.execute_cleanup_action(d)
         }
     }
 
     /// Core API: send message `m` to recipient `r`.
-    pub fn message<M: 'static + Send>(&mut self, r: &Arc<Ref<M>>, m: M) {
+    pub fn message<M: 'static + Send + std::fmt::Debug>(&mut self, r: &Arc<Ref<M>>, m: M) {
+        tracing::trace!(?r, ?m, "message");
         let r = Arc::clone(r);
         self.pending.queue_for(&r).push(Box::new(
-            move |t| t.with_entity(&r, |t, e| e.message(t, m))))
+            move |t| t.with_entity(&r, |t, e| {
+                tracing::trace!(?m, "delivered");
+                e.message(t, m)
+            })))
     }
 
     /// Core API: send message `m` to recipient `r`, which must be a
@@ -684,6 +723,21 @@ impl<'activation> Activation<'activation> {
             move |t| t.with_entity(&r, |t, e| e.sync(t, peer))))
     }
 
+    /// Registers the entity `r` in the list of stop actions for the active facet. If the facet
+    /// terminates cleanly, `r`'s [`stop`][Entity::stop] will be called.
+    ///
+    /// **Note.** If the actor crashes, the stop actions will *not* be called.
+    ///
+    /// Use [`RunningActor::add_exit_hook`] to install a callback that will be called at the
+    /// end of the lifetime of the *actor* rather than the facet. (Also, exit hooks are called
+    /// no matter whether actor termination was normal or abnormal.)
+    pub fn on_stop<M: 'static + Send>(&mut self, r: &Arc<Ref<M>>) {
+        if let Some(f) = self.active_facet() {
+            let r = Arc::clone(r);
+            f.stop_actions.push(Box::new(move |t| r.internal_with_entity(|e| e.stop(t))));
+        }
+    }
+
     /// Retrieve the [`Account`] against which actions are recorded.
     pub fn account(&self) -> &Arc<Account> {
         &self.pending.account
@@ -741,6 +795,7 @@ impl<'activation> Activation<'activation> {
         boot: F,
     ) {
         let mailbox = self.state.mailbox();
+        let facet = self.facet.clone();
         if let Some(f) = self.active_facet() {
             let token = CancellationToken::new();
             let task_id = NEXT_TASK_ID.fetch_add(BUMP_AMOUNT.into(), Ordering::Relaxed);
@@ -749,7 +804,7 @@ impl<'activation> Activation<'activation> {
                 let token = token.clone();
                 tokio::spawn(async move {
                     tracing::trace!(task_id, "linked task start");
-                    select! {
+                    let result = select! {
                         _ = token.cancelled() => {
                             tracing::trace!(task_id, "linked task cancelled");
                             Ok(())
@@ -768,13 +823,40 @@ impl<'activation> Activation<'activation> {
                             }
                             result
                         }
-                    }
+                    };
+                    let _ = facet.activate(
+                        Account::new(crate::name!("release_linked_task")), |t| {
+                            if let Some(f) = t.active_facet() {
+                                tracing::trace!(task_id, "cancellation token removed");
+                                f.linked_tasks.remove(&task_id);
+                            }
+                            Ok(())
+                        });
+                    result
                 }.instrument(name));
             }
             f.linked_tasks.insert(task_id, token);
         }
     }
 
+    /// Executes the given action after the given duration has elapsed (so long as the active
+    /// facet still exists at that time).
+    pub fn after(&mut self, duration: time::Duration, a: Action) {
+        self.at(time::Instant::now() + duration, a)
+    }
+
+    /// Executes the given action at the given instant (so long as the active facet still
+    /// exists at that time).
+    pub fn at<I: Into<tokio::time::Instant>>(&mut self, instant: I, a: Action) {
+        let facet = self.facet.clone();
+        let account = Arc::clone(self.account());
+        let instant = instant.into();
+        self.linked_task(crate::name!("Activation::at"), async move {
+            tokio::time::sleep_until(instant.into()).await;
+            facet.activate(account, a)
+        });
+    }
+
     fn enqueue_for_myself_at_commit(&mut self, action: Action) {
         let mailbox = self.state.mailbox();
         self.pending.queue_for_mailbox(&mailbox).push(action);
@@ -785,11 +867,14 @@ impl<'activation> Activation<'activation> {
         &mut self,
         name: tracing::Span,
         boot: F,
-    ) {
+    ) -> ActorRef {
+        let ac = Actor::new();
+        let ac_ref = ac.ac_ref.clone();
         self.enqueue_for_myself_at_commit(Box::new(move |_| {
-            Actor::new().boot(name, boot);
+            ac.boot(name, boot);
             Ok(())
         }));
+        ac_ref
     }
 
     /// Schedule the creation of a new actor when the Activation commits.
@@ -801,14 +886,17 @@ impl<'activation> Activation<'activation> {
         &mut self,
         name: tracing::Span,
         boot: F,
-    ) {
+    ) -> ActorRef {
+        let ac = Actor::new();
+        let ac_ref = ac.ac_ref.clone();
         let facet_id = self.facet.facet_id;
         self.enqueue_for_myself_at_commit(Box::new(move |t| {
             t.with_facet(true, facet_id, move |t| {
-                Actor::new().link(t).boot(name, boot);
+                ac.link(t).boot(name, boot);
                 Ok(())
             })
         }));
+        ac_ref
     }
 
     /// Create a new subfacet of the currently-active facet. Runs `boot` in the new facet's
@@ -843,7 +931,7 @@ impl<'activation> Activation<'activation> {
     /// its life, the Dataspace actor will have no outbound assertions, no child facets, and no
     /// linked tasks, so the only way to prevent it from being prematurely garbage collected is
     /// to use `prevent_inert_check` in its boot function.
-    pub fn prevent_inert_check(&mut self) -> Box<dyn FnOnce()> {
+    pub fn prevent_inert_check(&mut self) -> DisarmFn {
         if let Some(f) = self.active_facet() {
             Box::new(f.prevent_inert_check())
         } else {
@@ -908,10 +996,9 @@ impl<'activation> Activation<'activation> {
                         action(t)?;
                     }
                     let parent_facet_id = f.parent_facet_id;
-                    // if !alive, the drop will happen at the end of this function, but we
-                    // need it to happen right here so that child-facet cleanup-actions are
-                    // performed before parent-facet cleanup-actions.
-                    drop(f);
+                    f.retract_outbound(t);
+                    // ^ we need retraction to happen right here so that child-facet
+                    // cleanup-actions are performed before parent-facet cleanup-actions.
                     if let Some(p) = parent_facet_id {
                         if t.state.facet_exists_and_is_inert(p) {
                             t._terminate_facet(p, true)?;
@@ -919,6 +1006,8 @@ impl<'activation> Activation<'activation> {
                     } else {
                         t.state.shutdown();
                     }
+                } else {
+                    f.retract_outbound(t);
                 }
                 Ok(())
             })
@@ -961,13 +1050,18 @@ impl EventBuffer {
     }
 
     fn deliver(&mut self) {
+        tracing::trace!("EventBuffer::deliver");
         if !self.for_myself.is_empty() {
             panic!("Unprocessed for_myself events remain at deliver() time");
         }
         for (_actor_id, (tx, turn)) in std::mem::take(&mut self.queues).into_iter() {
-            if let Err(e) = send_actions(&tx, &self.account, turn) {
-                tracing::error!(actor=?_actor_id, ?e);
-            }
+            // Deliberately ignore send errors here: they indicate that the recipient is no
+            // longer alive. But we don't care about that case, since we have to be robust to
+            // crashes anyway. (When we were printing errors we saw here, an example of the
+            // problems it caused was a relay output_loop that received EPIPE causing the relay
+            // to crash, just as it was receiving thousands of messages a second, leading to
+            // many, many log reports of failed send_actions from the following line.)
+            let _ = send_actions(&tx, &self.account, turn);
         }
     }
 }
@@ -1107,6 +1201,7 @@ impl Actor {
             tx,
             mailbox: Weak::new(),
             exit_hooks: Vec::new(),
+            cleanup_actions: Map::new(),
             facet_nodes: Map::new(),
             facet_children: Map::new(),
             root: root.facet_id,
@@ -1198,7 +1293,8 @@ impl Actor {
                         if r.is_err() { return; }
                     }
                     SystemMessage::Crash(e) => {
-                        tracing::trace!(actor_id = ?self.ac_ref.actor_id, "SystemMessage::Crash({:?})", &e);
+                        tracing::trace!(actor_id = ?self.ac_ref.actor_id,
+                                        "SystemMessage::Crash({:?})", &e);
                         return terminate(Err(e));
                     }
                 }
@@ -1212,7 +1308,7 @@ impl Facet {
         Facet {
             facet_id: next_facet_id(),
             parent_facet_id,
-            cleanup_actions: Map::new(),
+            outbound_handles: Set::new(),
             stop_actions: Vec::new(),
             linked_tasks: Map::new(),
             inert_check_preventers: Arc::new(AtomicU64::new(0)),
@@ -1234,12 +1330,11 @@ impl Facet {
         }
     }
 
-    fn insert_retract_cleanup_action<M: 'static + Send>(&mut self, r: &Arc<Ref<M>>, handle: Handle) {
-        let r = Arc::clone(r);
-        self.cleanup_actions.insert(
-            handle,
-            CleanupAction::ForAnother(Arc::clone(&r.mailbox), Box::new(
-                move |t| t.with_entity(&r, |t, e| e.retract(t, handle)))));
+    fn retract_outbound(&mut self, t: &mut Activation) {
+        for handle in std::mem::take(&mut self.outbound_handles).into_iter() {
+            tracing::trace!(h = ?handle, "retract on termination");
+            t.retract(handle);
+        }
     }
 }
 
@@ -1330,13 +1425,31 @@ impl RunningActor {
         let no_kids = self.facet_children.get(&facet_id).map(|cs| cs.is_empty()).unwrap_or(true);
         if let Some(f) = self.get_facet(facet_id) {
             no_kids &&
-                f.cleanup_actions.is_empty() &&
+                f.outbound_handles.is_empty() &&
                 f.linked_tasks.is_empty() &&
                 f.inert_check_preventers.load(Ordering::Relaxed) == 0
         } else {
             false
         }
     }
+
+    fn insert_retract_cleanup_action<M: 'static + Send>(
+        &mut self,
+        r: &Arc<Ref<M>>,
+        handle: Handle,
+    ) {
+        let r = Arc::clone(r);
+        self.cleanup_actions.insert(
+            handle,
+            CleanupAction::ForAnother(Arc::clone(&r.mailbox), Box::new(
+                move |t| t.with_entity(&r, |t, e| {
+                    tracing::trace!(?handle, "retracted");
+                    if let Some(f) = t.active_facet() {
+                        f.outbound_handles.remove(&handle);
+                    }
+                    e.retract(t, handle)
+                }))));
+    }
 }
 
 impl Drop for Actor {
@@ -1352,13 +1465,8 @@ impl Drop for Facet {
             token.cancel();
         }
 
-        let to_clear = std::mem::take(&mut self.cleanup_actions);
-        {
-            let mut b = EventBuffer::new(Account::new(crate::name!("drop")));
-            for (_handle, r) in to_clear.into_iter() {
-                tracing::trace!(h = ?_handle, "retract on termination");
-                b.execute_cleanup_action(r);
-            }
+        if !self.outbound_handles.is_empty() {
+            panic!("Internal error: outbound_handles not empty at drop time");
         }
 
         tracing::trace!(facet_id = ?self.facet_id, "Facet::drop");
@@ -1558,6 +1666,9 @@ where
     fn sync(&mut self, t: &mut Activation, peer: Arc<Ref<Synced>>) -> ActorResult {
         t.with_entity(&self.underlying, |t, e| e.sync(t, peer))
     }
+    fn stop(&mut self, t: &mut Activation) -> ActorResult {
+        t.with_entity(&self.underlying, |t, e| e.stop(t))
+    }
     fn exit_hook(&mut self, t: &mut Activation, exit_status: &Arc<ActorResult>) -> ActorResult {
         self.underlying.internal_with_entity(|e| e.exit_hook(t, exit_status))
     }
